self.__precacheManifest = [
  {
    "revision": "cdf1bb69616000a770e7",
    "url": "./static/css/main.95ffe78b.chunk.css"
  },
  {
    "revision": "cdf1bb69616000a770e7",
    "url": "./static/js/main.e12d8782.chunk.js"
  },
  {
    "revision": "8c97409f0ee389fe75da",
    "url": "./static/js/runtime~main.d653cc00.js"
  },
  {
    "revision": "082027f6019b7a96bfcf",
    "url": "./static/css/2.4b811dd8.chunk.css"
  },
  {
    "revision": "082027f6019b7a96bfcf",
    "url": "./static/js/2.3723caf5.chunk.js"
  },
  {
    "revision": "74cdedb8050c1198d3c477e62d8bd849",
    "url": "./static/media/logo.74cdedb8.png"
  },
  {
    "revision": "588f6970a9445507e409207f97b54b90",
    "url": "./index.html"
  }
];